#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.

import copy
import warnings
from typing import List, Union
from .compartments import Compartment
from .polymer import OrderedMonomer, OrderedPolymer, MonomerCollection


class Species(OrderedMonomer):

    """ A formal species object for a CRN
     A Species must have a name. They may also have a material_type (such as DNA,
     RNA, Protein), and a list of attributes.
    """

    def __init__(self, name: str, material_type="", attributes: Union[List, None] = None,
                 compartment=None, **keywords):
        OrderedMonomer.__init__(self, **keywords)

        self.name = name
        self.material_type = material_type
        self._attributes = []  # Set this to avoid errors
        self.attributes = attributes
        self.compartment = compartment

    @property
    def attributes(self):
        if(not hasattr(self, "_attributes")):
            self._attributes = []
        return self._attributes

    @attributes.setter
    def attributes(self, attributes):
        if(not hasattr(self, "_attributes")):
            self._attributes = []
        if attributes is not None:
            if not isinstance(attributes, list):
                attributes = list(attributes)
            for attribute in attributes:
                self.add_attribute(attribute)
        elif attributes is None:
            self._attributes = []

    def remove_attribute(self, attribute: str):
        """
        removes an attribute from a Species
        """
        if(not hasattr(self, "_attributes") or attribute is None):
            return
        else:
            assert isinstance(attribute, str) and attribute.isalnum(), f"Attribute: {attribute} must be an alpha-numeric string"
            self._attributes = [a for a in self.attributes if a != attribute]

    def add_attribute(self, attribute: str):
        """
        Adds attributes to a Species
        """
        if(not hasattr(self, "_attributes")):
            self._attributes = []
        assert isinstance(attribute, str) and attribute is not None and attribute.isalnum(
        ), f"Attribute: {attribute} must be an alpha-numeric string"
        if attribute not in self.attributes:
            self._attributes.append(attribute)

    @property
    def name(self):
        if self._name is None:
            return ""
        else:
            return self._name

    @name.setter
    def name(self, name: str):
        if name is None:
            raise TypeError(f"Name must be a string. Recievied {name}")
        else:
            self._name = self._check_name(name)

    @property
    def compartment(self):
        return self._compartment

    @compartment.setter
    def compartment(self, compartment):
        if compartment is None:
            self._compartment = Compartment(name="default")
        else:
            if isinstance(compartment, str):
                self._compartment = Compartment(
                    name=self._check_name(compartment))
            elif isinstance(compartment, Compartment):
                self._compartment = compartment

    @property
    def direction(self):
        if hasattr(self, "_direction"):
            return self._direction
        else:
            return None
    
    @direction.setter
    def direction(self, direction):
        """
        This is inheritted from OrderedMonomer.
        A species with direction will use it as an attribute as well.
        This is overwritten to make direction an attribute
        """
        #Remove old direction from attributes
        self.remove_attribute(self.direction)
        #set the new direction
        self._direction = direction
        #Add it to attributes
        if direction is not None:
            self.add_attribute(direction)

    def remove(self):
        """
        Added functionality to remove direction as an attribute.
        """
        if self.direction is not None:
            self.remove_attribute(self.direction)
        return OrderedMonomer.remove(self)  # call the OrderedMonomer function

    # Note: this is used because properties can't be overwritten without setters being overwritten in subclasses.
    def _check_name(self, name):
        """
        Check that the string contains only underscores and alpha-numeric characters or is None. 
        Additionally cannot end in "_" or contain double "__", also cannot start with a number
        """
        if name is None:
            return name
        elif isinstance(name, str):
            no_underscore_string = name.replace("_", "")
            if no_underscore_string.isalnum() and "__" not in name and name[len(name)-1] != "_" and name[0].isalnum():
                return name
            else:
                raise ValueError(
                    f"name attribute {name} must consist of letters, numbers, or underscores and cannot contain double underscores or begin/end with a special character.")
        else:
            raise TypeError("Name must be a string.")

    @property
    def material_type(self):
        return self._material_type

    @material_type.setter
    def material_type(self, material_type: str):
        """ 
        Check that the string contains is alpha-numeric characters or "_" and that the first character is a letter. 
        If the name is a starts with a number, there must be a material type.
        """
        if material_type in [None, ""] and self.name[0].isnumeric():
            raise ValueError(
                f"species name: {self.name} contains a number as the first character and therefore requires a material_type.")
        elif material_type == None:
            self._material_type = None
        elif (material_type.replace("_", "").isalnum() and material_type.replace("_", "")[0].isalpha() and "__" not in material_type and material_type[len(material_type)-1] != "_") or material_type == "":
            self._material_type = material_type
        else:
            raise ValueError(
                f"material_type {material_type} must be alpha-numeric and start with a letter.")

    def __repr__(self):
        txt = ""
        if self.material_type not in ["", None]:
            txt = self.material_type + "_"

        txt += self.name

        if len(self.attributes) > 0 and self.attributes != []:
            for i in self.attributes:
                if i is not None:
                    txt += "_" + str(i)
        if self.compartment.name != 'default':
            # Only add a compartment name if it is not the default one.
            txt += "_" + self.compartment.name
        txt.replace("'", "")
        return txt

    def replace_species(self, species, new_species):
        if not isinstance(species, Species):
            raise ValueError(
                'species argument must be an instance of Species!')

        if not isinstance(new_species, Species):
            raise ValueError(
                'species argument must be an instance of Species!')

        if self == species:
            return new_species
        else:
            return self

    def get_species(self, **kwargs):
        """
        Used in some recursive calls where ComplexSpecies returns a list and Species will return just themselves (in a list)
        """
        return [self]

    def pretty_print(self, show_material=True, show_compartment=False, show_attributes=True,
                     show_initial_condition=False, **kwargs):
        """
        #A more powerful printing function. 
        Useful for understanding CRNs but does not return string identifiers.
        show_material toggles whether species.material is printed.
        show_attributes toggles whether species.attributes is printed
        """
        txt = ""
        if self.material_type not in ["", None] and show_material:
            txt = self.material_type + "["

        txt += self.name

        if self.compartment not in ["default", None] and show_compartment:
            txt += " in " + self.compartment.name + "."

        if len(self.attributes) > 0 and self.attributes != [] and show_attributes:
            txt += "("
            for i in self.attributes:
                if i is not None:
                    txt += str(i)+", "
            txt = txt[:-2]+")"

        txt.replace("'", "")
        if(hasattr(self, "direction") and self.direction is not None):
            txt += "-"+self.direction
        if self.material_type not in ["", None] and show_material:
            txt += "]"

        return txt

    def __eq__(self, other):
        """
        Overrides the default implementation
        Two species are equivalent if they have the same name, type, and attributes
        :param other: Species instance
        :return: boolean
        """

        #Note: "==" equality between parents and children can result in loops, so string equality is used
        if isinstance(other, Species) \
                and self.material_type == other.material_type \
                and self.name == other.name \
                and set(self.attributes) == set(other.attributes)\
                and str(self.parent) == str(other.parent)\
                and self.compartment == other.compartment\
                and self.position == other.position:
            return True
        else:
            return False

    def monomer_eq(self, other):
        """
        Same as normal equality, but does not check for parents or positions.
        """
        if isinstance(other, Species) \
                and self.material_type == other.material_type \
                and self.name == other.name \
                and set(self.attributes) == set(other.attributes)\
                and self.compartment == other.compartment:
            return True
        else:
            return False

    def __gt__(self, Species2):
        return self.name > Species2.name

    def __lt__(self, Species2):
        return self.name < Species2.name

    def __hash__(self):
        return str.__hash__(repr(self))

    def __contains__(self, item):
        return item in self.get_species()

    def contains_species_monomer(self, s):
        """Checks if the Species has a monomer (Species) inside of it, 
        but without checking Species.parent, Species.position, or direction. In effect, a
        less stringent version of __contains__."""
        s_copy = copy.deepcopy(s)
        s_copy.remove()
        for ss in self.get_species(recursive = True):
            ss_copy = copy.copy(ss)
            ss_copy.remove()
            if ss_copy == s_copy:
                return True
        return False

    @staticmethod
    def flatten_list(in_list) -> List:
        """Helper function to flatten lists
        """
        out_list = []
        if not isinstance(in_list, list):
            out_list.append(in_list)
        else:
            for element in in_list:
                if isinstance(element, list):
                    out_list += Species.flatten_list(element)
                elif(element is None):
                    pass
                else:
                    out_list += [element]
        return out_list


class WeightedSpecies:
    def __init__(self, species: Species, stoichiometry: int = 1):
        """Container object for a all types of species and its stoichiometry
        """
        self.species: Species = species
        self.stoichiometry: int = stoichiometry

    @property
    def stoichiometry(self):
        return self._stoichiometry

    @stoichiometry.setter
    def stoichiometry(self, new_stoichiometry):
        if new_stoichiometry <= 0:
            raise ValueError(
                f'Stoichiometry must be positive integer! We got {new_stoichiometry}!')
        self._stoichiometry = int(new_stoichiometry)

    def pretty_print(self, **kwargs):
        return f'{self.stoichiometry if self.stoichiometry > 1 else ""}{self.species.pretty_print(**kwargs)}'

    def replace_species(self, *args, **kwargs):
        return self.species.replace_species(*args, **kwargs)

    @staticmethod
    def _count_weighted_species(weighted_species: List):
        """Helper function merge the same species in a list with different stoichiometry

        >>> s1 = Species(name='a')
        >>> ws1 = WeightedSpecies(species=s1, stoichiometry=2)
        >>> ws2 = WeightedSpecies(species=s1, stoichiometry=5)
        >>> ws_list = [ws1, ws2]
        >>> freq_dict = WeightedSpecies._count_weighted_species(ws_list)
        >>> len(freq_dict)
        1

        :param weighted_species: list of weighted_species
        :return: unique list of weighted_species, i.e. set(weighted_species)
        """
        # convert to set doesn't work because we need only species equality
        unique_species = []
        for w_species in weighted_species:
            if not any(w_species.species == u_s.species for u_s in unique_species):
                unique_species.append(w_species)

        freq_dict = dict(zip(unique_species, [0]*len(unique_species)))
        for w_species in weighted_species:
            for key in freq_dict:
                if key.species == w_species.species:
                    freq_dict[key] += w_species.stoichiometry

        return freq_dict

    def __eq__(self, other):
        if other.__class__ is self.__class__:
            return (other.species, other.stoichiometry) == (self.species, self.stoichiometry)
        return False

    def __hash__(self):
        return hash(self.species)+hash(self.stoichiometry)


class ComplexSpecies(Species):
    """ 
    !!!ComplexSpecies and OrderedComplexSpecies should ALWAYS be created with the Complex function!!!

        A special kind of species which is formed as a complex of two or more species.
        Used for attribute inheritance and storing groups of bounds Species. 
        Note taht in a ComplexSpecies, the order of the species list does not matter.
        This means that ComplexSpecies([s1, s2]) = ComplexSpecies([s2, s1]). 
        This is good for modelling order-indpendent binding complexes.
        For a case where species order matters (e.g. polymers) use OrderedComplexSpecies
    """

    def __init__(self, species: List[Union[Species, str]], name: Union[str, None] = None, material_type="complex",
                 attributes=None, compartment=None, **keywords):

        # A little check to enforce use of Complex() to create ComplexSpecies
        if "called_from_complex" not in keywords or not keywords["called_from_complex"]:
            warnings.warn(
                "ComplexSpecies should be created using the Complex([List of Species]) function, not directly!")

        # Set species because it is used for default naming
        if len(Species.flatten_list(species)) <= 1:
            raise ValueError("chemical_reaction_network.complex requires 2 "
                             "or more species in its constructor.")
        self.species = species

        # call super class
        Species.__init__(self, name=name, material_type=material_type, attributes=attributes,
                         compartment=compartment)

    def __repr__(self):
        """
        ComplexSpecies add an additional "_" onto the end of their string representation
        This ensures that some edge cases are differentiated.
        """
        txt = Species.__repr__(self)
        txt += "_"
        return txt

    @property
    def name(self):
        if self._name is None:
            name = ""
            for s in self.species_set:
                count = self.species.count(s)
                name += str(s)+"_"
                if count > 1:
                    name += f"{count}x_"
            name = name[:-1]
            return name
        else:
            return self._name

    @name.setter
    def name(self, name: str):
        self._name = self._check_name(name)

    def __contains__(self, item):
        """
        Returns a list of species inside the ComplexSpecies
        """
        if not isinstance(item, Species):
            raise ValueError(
                "Operator 'in' requires chemical_reaction_network.Species (or a subclass). Received: "+str(item))
        if item in self.species:
            # this is the base case
            return True
        else:
            # this is the recursive part. We want to check all
            # internal complexes for the thing we're looking for
            for content in self.species:
                if isinstance(content, ComplexSpecies):
                    if item in content:
                        return True
            # if we got here then we've failed to find it
            return False

    @property
    def species_set(self):
        species_set = list(set(self.species))
        list.sort(species_set, key=lambda s: repr(s))
        return species_set

    @property
    def species(self):
        return self._species

    @species.setter
    def species(self, species):
        if not isinstance(species, list):
            raise TypeError(f"species must be a list: recieved {species}.")
        species = Species.flatten_list(species)
        if not all(isinstance(s, Species) for s in species):
            raise TypeError(
                f"recieved a non-species as a member of the list species: {species}.")
        else:
            list.sort(species, key=lambda s: repr(s))
            self._species = species

    def replace_species(self, species: Species, new_species: Species):
        """
        Replaces species with new_species in the entire Complex Species. Acts recursively on nested ComplexSpecies
        Does not act in place - returns a new ComplexSpecies. 
        """
        if not isinstance(species, Species):
            raise ValueError(
                'species argument must be an instance of Species!')

        if not isinstance(new_species, Species):
            raise ValueError(
                'species argument must be an instance of Species!')

        new_species_list = []
        for s in self.species:
            new_s = s.replace_species(species, new_species)
            new_species_list.append(new_s)

        new_name = None
        if self._name is not None:
            new_name = self.name

        return Complex(species=new_species_list, name=new_name, material_type=self.material_type, attributes=self.attributes)

    def get_species(self, recursive=False):
        """
        Returns all species in the ComplexSpecies. If recursive = True, returns species inside internal ComplexSpecies recursively as well.
        """
        if not recursive:
            species = [self]
        else:
            species = [self]
            for s in self.species:
                species += s.get_species(recursive=True)

        return species

    def pretty_print(self, show_material=True, show_compartment=False, show_attributes=True,
                     show_initial_condition=False, **kwargs):
        """
        A more powerful printing function. 
        Useful for understanding CRNs but does not return string identifiers.
        show_material toggles whether species.material is printed.
        show_attributes toggles whether species.attributes is printed
        """

        txt = ""
        if self.material_type not in ["", None] and show_material:
            txt += self.material_type
        txt += "["
        for s in self.species_set:
            count = self.species.count(s)
            if count > 1:
                txt += f"{count}x_"
            txt += s.pretty_print(show_material=show_material,
                                  show_attributes=False)+":"
        txt = txt[:-1]

        if self.compartment not in ["default", None] and show_compartment:
            txt += " in " + self.compartment.name + "."

        if len(self.attributes) > 0 and self.attributes != [] and show_attributes:
            txt += "("
            for i in self.attributes:
                if i is not None:
                    txt += str(i)+", "
            txt = txt[:-2]+")"

        txt.replace("'", "")
        if(hasattr(self, "direction") and self.direction is not None):
            txt += "-"+self.direction
        txt += "]"

        return txt

    def monomer_count(self, m):
        """
        Effectively self.species.count(m) using monomer_eq for equality
        """
        return sum([s.monomer_eq(m) for s in self.species])
    


class OrderedComplexSpecies(ComplexSpecies):
    """ 
    !!!ComplexSpecies and OrderedComplexSpecies should ALWAYS be created with the Complex function!!!

    A special kind of species which is formed as a complex of two or more species.
    In OrderedComplexSpecies the order in which the complex subspecies are is defined
    denote different species, eg [s1, s2, s3] != [s1, s3, s2].
    Used for attribute inheritance and storing groups of bounds Species. 
    """

    def __init__(self, species, name=None, material_type="ordered_complex", attributes=None,
                 compartment=None, **keywords):
        # Set species because it is used for default naming
        if len(Species.flatten_list(species)) <= 1:
            raise ValueError("chemical_reaction_network.complex requires 2 "
                             "or more species in its constructor.")
        self.species = species

        # Call the Species superclass constructor
        Species.__init__(self, name=name, material_type=material_type, attributes=attributes,
                         compartment=compartment)

    @property
    def name(self):
        if self._name is None:
            name = ""
            for s in self.species:
                if isinstance(s, str):
                    s = Species(name=s)
                if s.material_type not in [""]:
                    name += f"{s.material_type}_{s.name}_"
                else:
                    name += f"{s.name}_"
            name = name[:-1]
            return name
        else:
            return self._name

    @name.setter
    def name(self, name: str):
        self._name = self._check_name(name)

    @property
    def species(self):
        return self._species

    @species.setter
    def species(self, species):
        if not isinstance(species, list):
            raise TypeError(f"species must be a list: recieved {species}.")
        species = Species.flatten_list(species)
        if not all(isinstance(s, Species) for s in species):
            raise TypeError(
                f"recieved a non-species as a member of the list species: {species}.")
        else:
            self._species = species

    def replace_species(self, species: Species, new_species: Species):
        """
        Replaces species with new_species in the entire Complex Species. Acts recursively on nested ComplexSpecies
        """
        if not isinstance(species, Species):
            raise ValueError(
                'species argument must be an instance of Species!')

        if not isinstance(new_species, Species):
            raise ValueError(
                'species argument must be an instance of Species!')

        new_species_list = []
        for s in self.species:
            new_s = s.replace_species(species, new_species)
            new_species_list.append(new_s)

        new_name = None
        if self._name is not None:
            new_name = self.name

        return Complex(species=new_species_list, name=new_name, material_type=self.material_type, attributes=self.attributes, ordered=True)

    def pretty_print(self, show_material=True, show_compartment=False, show_attributes=True,
                     show_initial_condition=False, **kwargs):
        """
        A more powerful printing function. 
        Useful for understanding CRNs but does not return string identifiers.
        show_material toggles whether species.material is printed.
        show_attributes toggles whether species.attributes is printed
        """

        txt = ""
        if self.material_type not in ["", None] and show_material:
            txt += self.material_type

        txt += "["

        for s in self.species:
            txt += s.pretty_print(show_material=show_material,
                                  show_attributes=False)+":"
        txt = txt[:-1]

        if self.compartment not in ["default", None] and show_compartment:
            txt += " in " + self.compartment.name + "."

        if len(self.attributes) > 0 and self.attributes != [] and show_attributes:
            txt += "("
            for i in self.attributes:
                if i is not None:
                    txt += str(i)+", "
            txt = txt[:-2]+")"

        txt.replace("'", "")
        txt += "]"

        return txt


class OrderedPolymerSpecies(OrderedComplexSpecies, OrderedPolymer):
    """
    A class to represent OrderedPolymers which can also participate in chemical reactions.
    OrderedPolymerSpecies is made up of Species (which are also OrderedMonomers).

    The Species inside an OrderedPolymerSpecies are meant to model multiple binding sites and/or
    functional regions. ComplexSpecies can be formed inside an OrderedPolymer by passing
    the internal Species at a specific location.

    When used as an input to a reaction, OrderedPolymerSpecies can be passed 
    or one if its internal Species (eg a Species with Species.parent = OrderedPolymerSpecies)
    can also be used to produce the same reaction. This allows flexibility in the arguments to
    different Mechanisms. Sometimes, it is convenient to pass in the OrderedPolymerSpecies,
    sometimes it is convenient to pass an internal Species. Both will work from the point of view
    of any Mechanism.
    """
    default_material="ordered_polymer"
    def __init__(self,species, name=None, material_type = default_material, \
                   compartment=None, attributes = None,circular = False):

        self.material_type = material_type
        self.compartment = compartment
        self.parent = None
        self.position = None
        self.direction = None

        self.circular = circular

        if attributes is None:
            self.attributes = []
        else:
            self.attributes = attributes

        self._name = OrderedComplexSpecies._check_name(self, name)

        self.material_type = material_type
        #self.species = []
        monomers = []
        for specie in species:
            if isinstance(specie, OrderedPolymer) or isinstance(specie, PolymerConformation):
                raise NotImplementedError(
                    f"OrderedPolymer and PolymerConformation cannot be used as a Monomer at this time.")
            elif isinstance(specie, Species) and isinstance(specie, OrderedMonomer):
                monomers += [specie]
            elif (isinstance(specie, tuple) or isinstance(specie, list)) and (isinstance(specie[0], Species) and isinstance(specie[0], OrderedMonomer)):
                monomers += [specie]
            else:
                raise ValueError(
                    "{} should be a Species or list [Species, 'direction']".format(specie))
                # only species are acceptable

        OrderedPolymer.__init__(self, monomers)
        self.material_type = material_type

    @classmethod
    def from_polymer_species(cls, ops, replace_dict, **keywords):
        """
        Created a new OrderedPolymerSpecies with certain monomers replaced based upon replace_dict:

        inputs: replace_dict {monomer index --> new Species}
        outputs: OrderedPolymerSpecies
        """
        if(replace_dict == {}):
            #nothing to replace!
            return copy.deepcopy(ops)
        monomers = []
        for i in range(len(ops.polymer)):
            if i in replace_dict:
                monomers.append(replace_dict[i])
            else:
                monomer = copy.copy(ops[i])
                direction = monomer.direction
                monomer.remove()
                monomers.append([monomer, direction])

        #Set keywords
        if "circular" not in keywords:
            keywords["circular"] = ops.circular
        if "material_type" not in keywords:
            keywords["material_type"] = ops.default_material
        if "compartment" not in keywords:
            keywords["compartment"] = ops.compartment
        if "attributes" not in keywords:
            keywords["attributes"] = ops.attributes
        return cls(monomers, **keywords) #Produces a new OrderedPolymerSpecies

    @property
    def species_set(self):
        return set(self.polymer)

    @property
    def species(self):
        return self.polymer

    def get_species_list(self):
        return self.polymer

    @property
    def circular(self):
        if("circular" in self.attributes):
            return True
        else:
            return False

    @circular.setter
    def circular(self, value: bool):
        if(value):
            self.add_attribute("circular")
        else:
            self.remove_attribute("circular")

    def set_species_list(self, spec_tuple: tuple):
        OrderedPolymer.__init__(self, spec_tuple)

    @property
    def name(self):
        if self._name is None:
            name = ""
            for monomer in self.polymer:
                name+=str(monomer)+"_"
            name = name[:-1] #remove last underscore
        else:
            name = self._name
        
        return name

    def __hash__(self):
        ophash = OrderedPolymer.__hash__(self)
        ophash += hash(str(self))
        #hash(self.circular)+hash(self.name)+hash(self.material_type)+hash(self.attributes)
        return ophash

    def replace(self, position, part, direction=None):
        # TODO only change the name if the part we are replacing is actually different
        mydir = direction
        if((mydir is None) and (part.direction is not None)):
            mydir = part.direction
        if(part == self.polymer[position] and self.polymer[position].direction == mydir):
            # in this case we are replacing a part with the same thing, so do nothing
            # but it could be true that the reference changes? That shouldnt be
            pass
        else:
            OrderedPolymer.replace(
                self, position=position, part=part, direction=mydir)

    def __contains__(self, item):
        for part in self.species:
            part = copy.copy(part) #Only compare things which aren't None
            if item.parent is None:
                part.parent = None
            if item.direction is None:
                part.direciton = None
            if item.position is None:
                part.position = None
            if part == item:
                return True
        return False


class PolymerConformation(Species, MonomerCollection):
    """
    This class stores a set of PolymerSpecies and a set of connections between them in the form of ComplexSpecies containing Monomers inside the PolymerSpecies.

    The main function of this class is to provide a unique name to each conformation. The name is given by:

        conformation__[PolymerSpecies 1]_..._[PolymerSpecies N]_[ComplexSpecies_1 parent Polymer indices]_[ComplexSpecies_1]..._[ComplexSpecies_M]__
    
    where the list of PolymerSpecies and ComplexSpecies are in alphabetical order. 
    The ComplexSpecies parent Polymer indices notes which Polymers each Species in the ComplexSpecies comes from, with 'n' used for None.

    In general, users should not produce PolymerConformations directly. The Complex function will automatically produce these
    when a complex is formed involving Multiple OrderedMonomers contained within one or more PolymerSpecies.

    In effect, this can be thought of as a data structure for a hypergraph. The monomers of the PolymerSpecies are
    vertices and ComplexSpecies form edges that connect an arbitrary number of vertices (potentially including 
    other Species as well). Note that this class allows for multiple edges between the same sets of vertices.
    """

    def __init__(self, complexes = None, polymer = None, material_type = "conformation", name = None, **keywords):
        """
        complexes: a list of ComplexSpecies each of which must contain Monomers from the OrderedPolymerSpecies in the conformation
        """
        Species.__init__(self, name = name, material_type = material_type, **keywords)

        if isinstance(complexes, list) and len(complexes) == 0:
            complexes = None

        if (complexes is not None and polymer is not None) or (complexes is None and polymer is None):
            raise ValueError("PolymerConformation requires either: a list of ComplexSpecies which contain monomers from OrderedPolymerSpecies or a single OrderedPolymerSpecies in its constructor.")
        elif complexes is not None:
            self.complexes = complexes
        elif polymer is not None:
            #This order matters because complexes resets polymers
            self.complexes = []
            if isinstance(polymer, OrderedPolymerSpecies):
                self.polymers = [copy.deepcopy(polymer)]
            elif isinstance(polymer, list) and [isinstance(m, Species) for m in polymer]:
                self.polymers = [OrderedPolymerSpecies(polymer)]
            else:
                raise ValueError(f'polymer must be an OrderedPolymerSpecies or a list of Species. Received: {polymer}.')


    @classmethod
    def from_polymer_conformation(cls, pcs, complexes = None, complexes_to_remove = None, **keywords):
        """
        This function produces a new PolymerConformation from previously existing PolymerConformations and new Complexes.

        pcs: a list of PolymerConformations
        complexes: a list of complexes to add to the polymer conformation
        """

        if not isinstance(pcs, list) or not any([isinstance(pc, PolymerConformation) for pc in pcs]):
            raise TypeError(f"pcs must be a list of PolymerConformations. Recieved {pcs}.")

        if complexes is None:
            complexes = []
        if complexes_to_remove is None:
            complexes_to_remove = []

        #generate a list of all complexes
        for pc in pcs:
            for c in pc.complexes:
                if not any([c == cc for cc in complexes_to_remove]):
                    complexes.append(c)

        return cls(complexes, **keywords)

    @classmethod
    def from_polymer_replacement(cls, pc, old_polymers, new_polymers, **keywords):
        """
        This function produces a PolymerConformation from a previously existing PolymerConformation by replacing old_polymers with new_polymers
        
        pc: the PolymerConformation to replace polymers from.
        old_polymers: a list of PolymerSpecies instances. These must be the same instances stored inside pc or an error is thrown.
        new_polymers: a list of new PolymerSpecies instances to replace each of the old_polymers. Must be the same length as old_polymers.
        """
        if not isinstance(pc, PolymerConformation):
            raise TypeError(f"pc must be a PolymerConformation. Recieved {pc}.")

        if not isinstance(new_polymers, list) and all([isinstance(p, OrderedPolymerSpecies) for p in new_polymers]):
            raise TypeError(f"new_polymers must be a list of OrderedPolymerSpecies. Recieved: {new_polymers}")

        if not isinstance(old_polymers, list) and all([isinstance(p, OrderedPolymerSpecies) for p in old_polymers]):
            raise TypeError(f"new_polymers must be a list of OrderedPolymerSpecies. Recieved: {old_polymers}")

        if not all([any([p is pp for pp in pc.polymers]) for p in old_polymers]):
            raise ValueError("All OrderedPolymerSpecies in old_polymers must be contained (as instances, not string equivalents) in pc.polymers.")

        if len(old_polymers) != len(new_polymers):
            raise ValueError("old_polymers and new_polymers must be the same length.")

        complexes = pc.complexes
        new_complexes = []

        #Create a set of new ComplexeSpecies with the correct monomers inside
        for c in complexes:
            pol_inds = pc.get_polymer_indices(c)
            species = []
            for ci, s in enumerate(c.species):
                pi = pol_inds[ci]
                #If s is part of a polymer:
                if pi is not None:
                    #add the new_polymers[pi][s.position] to the new species
                    species.append(new_polymers[pi][s.position])
                #Otherwise, the species is not part of a polymer
                else:
                    species.append(s)

            #Create the new Complex
            if isinstance(c, OrderedComplexSpecies): #is the complex ordered?
                new_complexes += [OrderedComplexSpecies(species, material_type = c.material_type, attributes = c.attributes, comparment = c.compartment)]
            elif isinstance(c, ComplexSpecies):
                new_complexes += [ComplexSpecies(species, material_type = c.material_type, attributes = c.attributes, comparment = c.compartment)]
            else:
                raise TypeError(f"Invalid object found in PolymerConformation.complexes: {c}.")

        #Set keywords
        if "material_type" not in keywords:
            keywords["material_type"] = pc.material_type
        if "compartment" not in keywords:
            keywords["compartment"] = pc.compartment
        if "attributes" not in keywords:
            keywords["attributes"] = pc.attributes

        return cls(new_complexes, **keywords)

    def copy_remove_complexes(self, complexes):
        """
        Returns a new PolymerConformation without these complexes
        """

        if not isinstance(complexes, list):
            complexes = [complexes]

        #Check if the complexes are in the PolymerConformation
        for c in complexes:
            if c not in self.complexes:
                raise ValueError(f"Complex {c} not in PolymerConformation {self}.")

        new_complexes = [c for c in self.complexes if c not in complexes]

        return PolymerConformation(complexes = new_complexes)

    #To be a valid MonomerCollection
    @property
    def monomers(self):
        return tuple(self._species)

    @property
    def polymers(self):
        return self._polymers

    @polymers.setter
    def polymers(self, polymers):
        assert all([isinstance(p, OrderedPolymerSpecies) for p in polymers])
        polymers.sort()
        for p in polymers:
            p.parent = self
        self._polymers = polymers


    @property
    def complexes(self):
        return self._complexes


    @complexes.setter
    def complexes(self, complexes):
        """
        This setter copies the complexes and the polymers they connect into the PolymerConformation.
        This is done in such a way to preserve references between parents and children without relying on hash functions (in case two polymers are identical).
        """
        self._complexes = []
        self._polymers = []
        if not isinstance(complexes, list) or not all([isinstance(c, ComplexSpecies) for c in complexes]):
            raise TypeError(f"complexes must be a list containing ComplexSpecies. Recieved {complexes}.")

        complex_counts = [(sum([cc==c and cc.species == c.species for cc in complexes]), c) for c in complexes]
        if any([i[0] > 1 for i in complex_counts]):
            raise ValueError(f"Complexes contains two or more identical complexes. Recieved: {[c[1] for c in complex_counts if c[0]>1]}.")

        #Sort the complexes by their name, and the ids of the polymers inside of them to get a unique ordering for identically named Complexes and Polymers
        #this will produce a unique ordering of the internal polymers as well
        sort_func = lambda c: (c, tuple(id(s.parent) for s in c.species))
        complexes.sort(key = sort_func)


        polymers = [] #Original polymers stored here
        copied_polymers = [] #Polymers are copied here
        copied_complexes = [] #Complexes are copied here

        #Find all the polymers in the complexes. "is" is used instead of equality to deal with the possibility of
        #multiple instances (copies) of the same polymer being bound together.
        for c in complexes:
            c_copy = copy.deepcopy(c)
            complex_contains_polymer = False
            for i, s in enumerate(c.species):
                parent = s.parent

                if parent is not None and isinstance(parent, OrderedPolymerSpecies):
                    complex_contains_polymer = True
                    #parent has not been copied
                    if not any([parent is p for p in polymers]):
                        polymers.append(parent)
                        parent_copy = copy.deepcopy(parent)
                        parent_copy.parent = self #set the polymer's parent
                        copied_polymers.append(parent_copy)
                        c_copy.species[i] = parent_copy[s.position] #set s to be the new version from the deep copy
                    #parent has already been copied
                    else:
                        index = [parent is p for p in polymers].index(True) #get parent index
                        parent_copy = copied_polymers[index]
                        c_copy.species[i] = parent_copy[s.position] #set s to be the new version from the deep copy

                #This edgecase should not occur.
                elif parent is not None and not isinstance(parent, OrderedPolymerSpecies):
                    raise ValueError(f"Species {s} found inside complex {c} with a parent {parent} which is not an OrderedPolymerSpecies.")

            if not complex_contains_polymer:
                raise ValueError(f"Complex {c} does not contain any Species inside of Polymers.")

            #Set the parent of the ComplexSpecies to self
            c_copy.parent = self
            copied_complexes.append(c_copy)

        self.polymers = copied_polymers

        #Sort the complexes ordered by their polymer indices then name
        if len(copied_complexes)>0:
            complex_sort_func = lambda c: (tuple([i if i else -1 for i in self.get_polymer_indices(c)]), c)
            copied_complexes.sort(key = complex_sort_func)
        self._complexes = copied_complexes

    
    def get_polymer_indices(self, c):
        #Takes a complex and returns a list of indices to the polymers that complex contains monomers from
        #this complex should be in self.complexes. 
        #If a Species in the Complex is not in a Polymer, None is added to the list.
        indices = []
        for j, s in enumerate(c.species):
            parent = s.parent
            for i, p in enumerate(self.polymers):
                if parent is p:
                    indices.append(i)
            if len(indices) <= j:
                indices.append(None) 

        return indices

    def get_polymer_positions(self, c, polymer_ind):
        #Takes a complex and the index of a polymer in the conformation and returns a list of positions that ComplexSpecies is bound at
        p = self.polymers[polymer_ind]
        positions = []
        for s in c.species:
            if s.parent is p:
                positions.append(s.position)
            else:
                positions.append(None)
        return tuple(positions)

    def get_polymer(self, p):
        #Takes a polymer and returns a matching instance inside self.polymers (or None)
        p = copy.copy(p)
        p.parent = self
        if p in self.polymers:
            i = self.polymers.index(p)
            return self.polymers[i]
        else:
            return None

    def count_polymer(self, p):
        p = copy.copy(p)
        p.parent = self
        count = 0
        for pp in self.polymers:
            if p == pp:
                count += 1
        return count

    def get_complex(self, c):
        #takes a ComplexSpecies and returns the instance of this Species in the PolymerConformation (or None)
        c = copy.copy(c)
        c.parent = self
        #if c in self.complexes:
        #    i = self.complexes.index(c)
        #    return self.complexes[i]
        for cc in self.complexes:
            if cc == c and [(s.position, str(s.parent)) for s in c.species] == [(s.position, str(s.parent)) for s in cc.species]:
                return cc
        return None

    def remove_complex(self, c):
        #Removes a copy of this complex from PolymerConformation if possible
        if c in self.complexes:
            self.complexes.remove(c)
        else:
            raise ValueError(f"Complex {c} not in PolymerConformation {self}.")

    def get_complexes_at(self, p_ind, p_pos):
        #returns all complexes in polymer p_ind bound to location p_pos
        found_complexes = []
        for c in self.complexes:
            for s in c.species:
                if s.parent == self.polymers[p_ind] and s.position == p_pos:
                    found_complexes.append(c)
        return found_complexes

    @property
    def name(self):
        if self._name is None:
            #If there is nothing in the PolymerConformation, use the name of the internal OrderedPolymerSpecies
            if len(self.complexes) == 0 and len(self.polymers) == 1:
                return str(self.polymers[0])
            else:

                name = ""
                for p in self.polymers:
                    name += "_"+str(p)
                for c in self.complexes:
                    parent_inds = self.get_polymer_indices(c)
                    name += "_"
                    for i, ind in enumerate(parent_inds):
                        if ind is None:
                            name+="n"
                        else:
                            locs = list(self.get_polymer_positions(c, ind))
                            loc = locs[i]

                            name+=f"p{ind}l{loc}"
                    name += "_"+str(c)
                return name
        else:
            return self._name

    @name.setter
    def name(self, name: str):
        self._name = self._check_name(name)

    def __repr__(self):
        """
        PolymerConformations add an additional "_" onto the end of their string representation
        This ensures that some edge cases are differentiated.

        A PolymerConformation with no attributes or complexes consisting of just one OrderedPolymerSpecies uses the same representation as the OrderedPolymerSpecies.
        """
        if len(self.complexes) == 0 and len(self.polymers) == 1:
            txt = Species.__repr__(self)
            txt = txt.replace("conformation_", "")
        else: 
            txt = Species.__repr__(self)
            txt += "_"
        return txt


class Complex:
    """
     !!!ComplexSpecies and OrderedComplexSpecies should ALWAYS be created with the Complex function!!!

    Complex is not a class that gets instantiated - it creates ComplexSpecies and OrderedComplexSpecies.
    The Logic encoded in the __new__ function is used to insert these classes into the binding sites of
    OrderedPolymerSpecies.

    arguments:
    species: a list of species to put into ComplexSpecies or OrderedComplexSpecies

    keywords:
    ordered: whether to produce an OrderedComplexSpecies (default = False)
    """
    def __new__(cls, *args, **keywords):
        """
        This function effectively produces the instance of the correct Species Class based upon the arguments passed in.

        Cases: Here species refer to the Species in the Species list passed into the construct.
        1.  No Species have parents.
            Produces: an ComplexSpecies or an OrderedComplexSepcies
        2.  A single Species S has a parent which is an OrderedPolymerSpecies with no parent. 
            Produces: an OrderedPolymerSpecies with a ComplexSpecies or OrderedComplexSpecies containing S in S's location in the OrderedPolymerSpecies.
        3.  [Error Case] Multiple Species S have parents which are OrderedPolymerSpecies without parents.
        4.  [Error Case] Entire OrderedPolymerSpecies inside PolymerConformations are being Complexed Together
        5.  One or More Species S have parents which are OrderedPolymerSpecies with parents and/or PolymerConformations
            Produces: a (Ordered)ComplexSpecies containing all S inside a PolymerConformation which merges all PolymerComformation Complexes.
        """
        species = []
        # below is extracting the "species" keywords from the args
        keywarg = None
        if("species" in keywords):
            keywarg = True
            species = keywords["species"]
            del keywords["species"]
        elif(len(args) >= 1):
            keywarg = False
            species = args[0]
            args = args[1:]

        if not isinstance(species, list):
            raise TypeError(f"First argument to Complex (or species keyword), must be a list of Species; recieved {species}.")

        # Check whether ot make a ComplexSpecies or OrderedComplexSpecies
        if "ordered" in keywords:
            ordered = keywords["ordered"]
            del keywords["ordered"]
            ComplexClass = OrderedComplexSpecies
        else:
            ordered = False
            ComplexClass = ComplexSpecies

        #Use to supress errors in ComplexSpecies and OrderedComplexSpecies
        keywords["called_from_complex"] = True

        parent_species = []  # parent_species is a list of OrderedPolymerSpecies and/or PolymerConformations
        bindlocs = []  # bindloc is the location a Species is bound to parent_species
        insertlocs = [] #insertloc is the location of the species inside the Complex.species list (important to maintain order in OrderedComplexSpecies)
        child_species = [] #the species with parents
        other_species = []  # Other species in the Complex

        # Below cycle through species and see if one has a parent. If it does, that means the species is
        # in an OrderedPolymerSpecies and the Complex should be formed around it.
        for i, specie in enumerate(species):
            if(hasattr(specie, "parent") and (specie.parent is not None)):
                #This adds to a growing list of parents, which will be placed in an PolymerConformation
                # It is very important to deepcopy here because the underlying OrderedPolymerSpecies or PolymerConformation will be modified.
                # insertlocs stores the order of the species for creating OrderedComplexSpecies
                parent_species.append(specie.parent)
                bindlocs.append(specie.position)
                insertlocs.append(i)
                child_species.append(specie)
            else:
                other_species += [specie]

        #Case 1: If no OrderedPolymerSpecies is found, just call the regular constructor.
        if len(parent_species) == 0:
            return ComplexClass(species, *args, **keywords)

        #Case 2: the Complex is being formed inside an OrderedPolymerSpecies (which is not in a PolymerConformation).
        elif len(parent_species) == 1 and isinstance(parent_species[0], OrderedPolymerSpecies) and (parent_species[0].parent is None):
            parent_species = parent_species[0]
            bindloc = bindlocs[0]

            # Create an OrderedcomplexSepcies or ComplexSpecies
            child = copy.copy(child_species[0])
            child_direction = child_species[0].direction
            child.remove() #Remove the child species, the new_complex will replace it
            species.remove(child_species[0])
            species.insert(insertlocs[0], child) #place the new child in the list in the proper location to preserve order
            new_complex = ComplexClass(species, *args, **keywords) #create the Complex

            #Create a new OrderedPolymerSpecied which is copied from the parent with the new complex replacing bindloc (inheriting the same direction).
            new_polymer_species = OrderedPolymerSpecies.from_polymer_species(parent_species, {bindloc:[new_complex, child_direction]})


            #Case 2: OrderedPolymerSpecies has no parent
            if parent_species.parent is None:
                return new_polymer_species[bindloc]

            #(Old Case 3 now case 4) OrderedPolymerSpecies is inside a PolymerConformation has been moved to case 
            #elif isinstance(parent_species.parent, PolymerConformation):
            #    raise RuntimeError("This case should no longer occur as defined")
                #create a new PolymerConformation that replaces the appropriate monomer in the polymer
            #    new_pc = PolymerConformation.from_polymer_replacement(parent_species.parent, [parent_species], [new_polymer_species])
            #    return new_pc.get_polymer(new_polymer_species)[bindloc] #return the newly created Monomer attached to the new OrderedPolymerSpecies inside the new PolymerConformation
            #else:
            #    #This error should never occur
            #    raise TypeError(f"Unknown parent type {type(parent_species)} recieved for {parent_species} .parent {parent_species.parent}.")
        
        #Case 3-4: In the following cases, multiple species may have parents
        elif all([isinstance(p, OrderedPolymerSpecies) for p in parent_species]) and any([p.parent is None for p in parent_species]):
            raise TypeError("In order to form Complexes between Monomers inside OrderedPolymerSpecies, each OrderedPolymerSpecies must be placed in one or more PolymerConformations.")

        elif any([isinstance(s, OrderedPolymerSpecies) and s.parent is not None for s in species]):
            raise ValueError("OrderedPolymerSpecies cannot be combined into a Complex if they are already part of an PolymerConformation. Maybe you meant to Complex specific monomers?")
        
        #Case 5: Multiple species in one more more PolymerConformations are being Complexed Together
        else:
            pcs = []
            merged_species = other_species #these Species will go inside the ComplexSpecies later

            #Cycle through Species with parents
            insert_loc_offset = 0
            for i, p in enumerate(parent_species):
                s = child_species[i]

                #if the parent is a OrderedPolymerSpecies
                if isinstance(p, OrderedPolymerSpecies):
                    merged_species.insert(insertlocs[i]+insert_loc_offset, s)

                    #if the Polymer is already in a Conformation...
                    if p.parent is not None and not any([p.parent is pp for pp in pcs]):
                        pcs.append(p.parent)

                #if the parent is a PolymerConformation and child is a ComplexSpecies
                elif isinstance(p, PolymerConformation) and isinstance(s, ComplexSpecies) and not isinstance(s, OrderedPolymerSpecies):
                    #Store all the unique PolymerConformations
                    if not any([p is pp for pp in pcs]):
                        p = copy.copy(p)
                        p.remove_complex(s)
                        pcs.append(p)

                    #Merge the species lists
                    merged_species = merged_species[:insertlocs[i]+insert_loc_offset]+s.species+merged_species[insertlocs[i]+insert_loc_offset:]
                    insert_loc_offset += len(s.species)-1 #this takes care of ordering offsets during the merge
                else:
                    raise TypeError(f"Cannot form a complex from {species}. Invalid Parent Species {p} for child {s}.")

            # Create a Complex and merged PolymerConformation
            new_complex = ComplexClass(merged_species, *args, **keywords)
            new_pc = PolymerConformation.from_polymer_conformation(pcs, [new_complex])
            return new_pc.get_complex(new_complex)




#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.


from .propensities import (HillNegative, HillPositive, MassAction, Propensity,
                           ProportionalHillNegative, ProportionalHillPositive)

from.species import *
import copy
import itertools
from typing import List, Union
from warnings import warn

from .utils import remove_bindloc


class Reaction(object):
    """An abstract representation of a chemical reaction in a CRN.

    A reaction has the form:
    .. math::
       \sum_i n_i I_i --> \sum_i m_i O_i @ rate = k
       where n_i is the count of the ith input, I_i, and m_i is the count of the
       ith output, O_i.
    If the reaction is reversible, the reverse reaction is also included:
    .. math::
       \sum_i m_i O_i  --> \sum_i n_i I_i @ rate = k_rev
    """
    def __init__(self, inputs: Union[List[Species], List[WeightedSpecies]],
                 outputs: Union[List[Species], List[WeightedSpecies]],
                 propensity_type: Propensity):

        if len(inputs) == 0 and len(outputs) == 0:
            warn("Reaction Inputs and Outputs both contain 0 Species.")

        self.inputs = remove_bindloc(Species.flatten_list(inputs))
        self.outputs = remove_bindloc(Species.flatten_list(outputs))
        self.propensity_type = propensity_type

    @property
    def propensity_type(self) -> Propensity:
        return self._propensity_type

    @propensity_type.setter
    def propensity_type(self, new_propensity_type: Propensity):
        """Replace the propensity type associated with the reaction object.

        :param new_propensity_type: Valid propensity type
        """
        if not Propensity.is_valid_propensity(new_propensity_type):
            raise ValueError(f'unknown propensity type: {new_propensity_type} '
                             f'({type(new_propensity_type)})!')

        self._propensity_type = new_propensity_type

    @classmethod
    def from_massaction(cls, inputs: Union[List[Species], List[WeightedSpecies]],
                        outputs: Union[List[Species], List[WeightedSpecies]],
                        k_forward: float, k_reverse: float = None):
        """Initialize a Reaction object with mass action kinetics.

        :param inputs:
        :param outputs:
        :param k_forward:
        :param k_reverse:
        :return: Reaction object
        """
        mak = MassAction(k_forward=k_forward, k_reverse=k_reverse)

        return cls(inputs=inputs, outputs=outputs, propensity_type=mak)

    @property
    def is_reversible(self) -> bool:
        return self.propensity_type.is_reversible

    @property
    def inputs(self) -> List[WeightedSpecies]:
        return self._input_complexes

    @inputs.setter
    def inputs(self, new_input_complexes: List[WeightedSpecies]):
        self._input_complexes = Reaction._check_and_convert_complex_list(complexes=new_input_complexes)

    @property
    def outputs(self) -> List[WeightedSpecies]:
        return self._output_complexes

    @outputs.setter
    def outputs(self, new_output_complexes: List[WeightedSpecies]):
        self._output_complexes = Reaction._check_and_convert_complex_list(complexes=new_output_complexes)

    @staticmethod
    def _check_and_convert_complex_list(complexes: Union[List[Species], List[WeightedSpecies]]) -> List[WeightedSpecies]:
        if all([isinstance(one_complex, Species) for one_complex in complexes]):
            # we wrap each Species object to WeightedSpecies
            complexes = [WeightedSpecies(species=species) for species in complexes]
        else:
            if not all([isinstance(one_complex, WeightedSpecies) for one_complex in complexes]):
                raise TypeError(f'inputs must be list of Species or list of ChemicalComplexes! Recieved {complexes}')

        # filter out duplicates and adjust stoichiometry
        out_list = []
        # Create a dictionary of unique species and their stoichiometry count
        stoichiometry_count = WeightedSpecies._count_weighted_species(complexes)
        for one_complex, stoichiometry in stoichiometry_count.items():
            new_complex = WeightedSpecies(species=one_complex.species, stoichiometry=stoichiometry)
            out_list.append(new_complex)

        return out_list

    #@property
    #def k_forward(self):
    #    return self.propensity_type.k_forward

    #@property
    #def k_reverse(self):
    #    return self.propensity_type.k_reverse

    def replace_species(self, species: Species, new_species: Species):
        """Replaces species with new_species in the reaction.

        :param species: species to be replaced
        :param new_species: the new species the old species is replaced with
        :return: a new Reaction instance
        """
        if not isinstance(species, Species) or not isinstance(new_species, Species):
            raise ValueError('both species and new_species argument must be an instance of Species!')

        new_inputs = []
        for s in self.inputs:
            new_s = s.replace_species(species, new_species)
            new_inputs.append(new_s)

        new_outputs = []
        for s in self.outputs:
            new_s = s.replace_species(species, new_species)
            new_outputs.append(new_s)

        # get a shallow copy of the parameters and species, so we can replace some of them
        propensity_type_dict = copy.copy(self.propensity_type.propensity_dict)
        for key, prop_species in propensity_type_dict['species'].items():
            propensity_type_dict['species'][key] = prop_species.replace_species(species, new_species)

        new_propensity_type = self.propensity_type.from_dict(propensity_type_dict)

        new_r = Reaction(inputs=new_inputs, outputs=new_outputs, propensity_type=new_propensity_type)
        return new_r

    def __repr__(self):
        """Helper function to print the text of a rate function"""
        return self.pretty_print(show_rates=False, show_material=True, show_attributes=True, show_parameters = False)

    def pretty_print(self, show_rates=True, show_material=True, show_attributes=True, show_parameters = True, **kwargs):

        kwargs['show_rates'] = show_rates
        kwargs['show_material'] = show_material
        kwargs['show_attributes'] = show_attributes

        txt = '+'.join([s.pretty_print(**kwargs) for s in self.inputs])

        if self.is_reversible:
            txt += " <--> "
        else:
            txt += " --> "

        txt += '+'.join([s.pretty_print(**kwargs) for s in self.outputs])
        if show_rates:

            #These kwargs are essential for massaction propensities
            kwargs["reaction"] = self
            if "stochastic" not in kwargs:
                kwargs["stochastic"] = False

            txt += "\n"+f'{self.propensity_type.pretty_print(**kwargs)}'

        return txt

    def __eq__(self, other):
        """Two reactions are equivalent if they have the same inputs, outputs,
           and propensity."""
        if not isinstance(other, Reaction):
            raise TypeError(f'Only reactions can be compared with reaction! We got {type(other)}.')

        if len(self.inputs) != len(other.inputs) or len(self.outputs) != len(other.outputs):
            return False

        return (set(self.inputs), set(self.outputs), self.propensity_type) == (set(other.inputs), set(other.outputs), other.propensity_type)

    def __contains__(self, item: Species):
        """It checks whether a species is part of a reaction.

         it checks the input and output lists as well as the propensity type for the species

        :param item: a Species instance
        :return: bool
        :exception NotImplementedError for non-Species objects
        """
        if isinstance(item, Species):
            if item in self.inputs \
                    or item in self.outputs \
                    or item in self.propensity_type.species:
                return True
        else:
            raise NotImplementedError
        return False

    @property
    def species(self) -> List[Species]:
        """returns a list of species in the reactions collected from the inputs
        and outputs and the propensity (e.g. Hill kinetics has species in it).

        :return: list of species in the reactions
        """
        in_part = []
        for s in self.inputs:
            in_part.extend(Species.flatten_list(s.species))
        out_part = []
        for s in self.outputs:
            out_part.extend(Species.flatten_list(s.species))

        return list(itertools.chain(in_part, out_part, self.propensity_type.species))




#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.

import copy
import itertools as it
from warnings import warn

from .dna_part import DNA_part
from .mechanisms_binding import (Combinatorial_Cooperative_Binding,
                                 One_Step_Cooperative_Binding)
from .mechanisms_txtl import (NegativeHillTranscription,
                              PositiveHillTranscription)
from .species import Species
from .components_basic import DNA, RNA
from .utils import remove_bindloc


#TODO put remove_bindloc in the component instead of in dna_construct
class Promoter(DNA_part):
    """A basic Promoter class with no regulation. Needs to be included in a DNAassembly or DNAconstruct to function.
    """
    def __init__(self, name, assembly=None,
                 transcript=None, length=0,
                 mechanisms=None, parameters=None, protein=None,dna_to_bind=None, **keywords):
        self._dna_bind = dna_to_bind
        self.length = length

        if transcript is None and assembly is None:
            self.transcript = None
        elif transcript is None:
            self.transcript = Species(assembly.name, material_type="rna")
        elif type(transcript) == list:
            self.transcript = transcript
        else:
            self.transcript = self.set_species(transcript, material_type='rna')
        
        if(isinstance(protein,str)):
            self.protein = [self.set_species(protein,material_type="protein")]
        elif(isinstance(protein,Species)):
            self.protein = [protein]
        elif(isinstance(protein,list)):
            self.protein = protein
        else:
            self.protein = None
        #Promoter should not have initial conditions. These need to be in DNAAssembly or DNAConstruct
        if "initial_concentration" in keywords.values() and keywords["initial_concentration"] is not None:
            raise AttributeError("Cannot set initial_concentration of a Promoter. Must set initial_concentration for the DNAassembly or DNAConstruct.")
        if "initial_condition_dictionary" in keywords.values() and keywords["initial_condition_dictionary"] is not None:
            raise AttributeError("Cannot set initial_condition_dictionary of a Promoter. Must set initial_condition_dictionary for the DNAassembly or DNAconstruct.")

        DNA_part.__init__(self, name = name, mechanisms = mechanisms,
                           parameters = parameters, assembly=assembly, **keywords)

    def update_species(self):
        mech_tx = self.get_mechanism("transcription")
        species = []

        species += mech_tx.update_species(dna = self.dna_to_bind, \
            transcript = self.transcript, protein = self.get_protein_for_expression(),
            component = self, part_id = self.name)

        return species

    @property
    def dna_to_bind(self):
        if(self._dna_bind is None):
            if(self.assembly is None):
                return None
            else:
                self._dna_bind = self.assembly.dna
        return self._dna_bind

    @dna_to_bind.setter
    def dna_to_bind(self,value):
        self._dna_bind = value
        
    def get_species(self):
        return None

    def update_reactions(self):
        mech_tx = self.get_mechanism("transcription")
        reactions = []

        reactions += mech_tx.update_reactions(dna = self.dna_to_bind, \
                        component = self, part_id = self.name, complex = None,
                        transcript = self.transcript, protein = self.get_protein_for_expression())
        return reactions

    def update_component(self,internal_species=None,**keywords):
        """returns a copy of this component, except with the proper fields updated"""
        if(isinstance(self.parent,RNA)):
            return None
        elif(isinstance(self.parent,DNA)):
            out_component = copy.copy(self)
            out_component.dna_to_bind = internal_species
            return out_component
        else:
            raise TypeError(f"Unknown parent class {type(self.parent)}, expect either DNA_construct or RNA_construct")


    #Used for expression mixtures where transcripts are replaced by proteins
    def get_protein_for_expression(self):
        if self.transcript is None:
            return self.protein
        else:
            return None

    @classmethod
    def from_promoter(cls, name, assembly, transcript, protein):
        """Helper function to initialize a promoter instance from another promoter or str.

        :param name: either string or an other promoter instance
        :param assembly:
        :param transcript:
        :param protein:
        :return: Promoter instance
        """
        if isinstance(name, Promoter):
            promoter_instance = copy.deepcopy(name)
            promoter_instance.assembly = assembly
            promoter_instance.transcript = transcript
            promoter_instance.protein = protein
        elif isinstance(name, str):
            promoter_instance = cls(name=name, assembly=assembly,
                                    transcript=transcript, protein=protein)
        else:
            raise TypeError(f'Promoter can be initialized from string or another promoter! We got {type(name)}')
        return promoter_instance


class RegulatedPromoter(Promoter):
    """
    A Promoter class with simple regulation.
    regulators = [list of species]
    Each regulator binds independently to the Promoter to regulate it.
    """
    def __init__(self, name: str, regulators, leak=True, assembly=None,
                 transcript=None, length=0, mechanisms=None,
                 parameters=None , **keywords):

        Promoter.__init__(self, name = name, assembly = assembly,
                          transcript = transcript, length = length,
                          mechanisms = mechanisms, parameters = parameters,
                          **keywords)

        if not isinstance(regulators, list):
            regulators = [regulators]

        self.regulators = []
        for regulator in regulators:
            self.regulators += [self.set_species(regulator, material_type = "protein")]

        self.leak = leak

        self.add_mechanism(One_Step_Cooperative_Binding(), "binding")
        self.complexes = []
        
    def update_species(self):
        mech_tx = self.get_mechanism("transcription")
        mech_b = self.get_mechanism('binding')
        species = []

        
        self.complexes = []
        if self.leak is not False:
            species += mech_tx.update_species(dna = self.dna_to_bind, transcript = self.transcript,\
                             component = self, part_id = self.name+"_leak", protein = self.get_protein_for_expression())

        for i in range(len(self.regulators)):
            regulator = self.regulators[i]

            species_b = mech_b.update_species(regulator, self.dna_to_bind, part_id = self.name+"_"+regulator.name,\
                             component = self, protein = self.get_protein_for_expression())
            species += species_b

            #Find complexes containing DNA and the regulator
            #DNA should be *not* a part of an OrderedPolymer for this to work
            #dna_simple = copy.deepcopy(self.dna_to_bind)
            #dna_simple.remove()
            for s in species_b:
                if s.contains_species_monomer(self.dna_to_bind) and s.contains_species_monomer(regulator):
                    self.complexes += [s]

                    species += mech_tx.update_species(dna = s, transcript = self.transcript, \
                            protein = self.get_protein_for_expression(), part_id = self.name+"_"+regulator.name, component = self)
        return species

    def update_reactions(self):
        reactions = []
        mech_tx = self.get_mechanism("transcription")
        mech_b = self.get_mechanism('binding')

        if self.leak is not False:
            reactions += mech_tx.update_reactions(dna = self.dna_to_bind, component = self, part_id = self.name+"_leak", \
                                                            transcript = self.transcript, protein = self.get_protein_for_expression())

        for i in range(len(self.regulators)):
            regulator = self.regulators[i]
            complex_ = self.complexes[i]

            reactions += mech_b.update_reactions(regulator, self.dna_to_bind, component = self, \
                                                                    part_id = self.name+"_"+regulator.name,protein=self.protein)
            
            reactions += mech_tx.update_reactions(dna = complex_, component = self, part_id = self.name+"_"+regulator.name, \
                                            transcript = self.transcript, protein = self.get_protein_for_expression())

        return reactions


class ActivatablePromoter(Promoter):
    """
    A class for a promoter which can be activated by a single species, modelled as a positive hill function
    """
    def __init__(self, name, activator, transcript = None, leak = False, **keywords):
        #Always call the superclass __init__() with **keywords passed through
        Promoter.__init__(self, name = name, transcript = transcript, **keywords)

        #Set the Regulator
        #Component.set_species(species, material_type = None, attributes = None)
        # is a helper function that allows the input to be a Species, string, or Component.
        self.activator = self.set_species(activator) 

        self.leak = leak #toggles whether or not there is a leak reaction
        
        #Non-default Mechanisms are added to the Component with .add_mechanism
        self.add_mechanism(PositiveHillTranscription(), "transcription", overwrite = True)
        

    def update_species(self, **keywords):
        #Mechanisms are stored in an automatically created dictionary: mechanism_type --> Mechanism Instance.
        mech_tx = self.get_mechanism("transcription")
        
        species = [] #A list of species must be returned
        species += mech_tx.update_species(dna = self.dna_to_bind, transcript = self.transcript, regulator = self.activator, 
            part_id = self.name+"_"+self.activator.name, leak = self.leak, component = self, protein = self.get_protein_for_expression())
        
        return species

    def update_reactions(self, **keywords):
        mech_tx = self.get_mechanism("transcription")
        
        reactions = [] #a list of reactions must be returned


        reactions += mech_tx.update_reactions(dna = self.dna_to_bind, transcript = self.transcript, regulator = self.activator, 
                                             component = self, part_id = self.name+"_"+self.activator.name, leak = self.leak, protein = self.get_protein_for_expression())

        
        return reactions

class RepressiblePromoter(Promoter):
    """
    A class for a promoter which can be repressed by a single species, modelled as a negative hill function
    """
    def __init__(self, name, repressor, transcript = None, leak = False, **keywords):
        #Always call the superclass __init__() with **keywords passed through
        Promoter.__init__(self, name = name, transcript = transcript, **keywords)

        #Set the Regulator
        #Component.set_species(species, material_type = None, attributes = None)
        # is a helper function that allows the input to be a Species, string, or Component.
        self.repressor = self.set_species(repressor) 

        self.leak = leak #toggles whether or not there is a leak reaction
        
        #Mechanisms are inherited from the Mixture unless set specifically in self.default_mechanisms.
        self.add_mechanism(NegativeHillTranscription(), "transcription", overwrite = True)
        
        

    def update_species(self, **keywords):
        #Mechanisms are stored in an automatically created dictionary: mechanism_type --> Mechanism Instance.
        mech_tx = self.get_mechanism("transcription")
        
        species = [] #A list of species must be returned
        species += mech_tx.update_species(dna = self.dna_to_bind, transcript = self.transcript, regulator = self.repressor, component = self, 
            part_id = self.name+"_"+self.repressor.name, leak = self.leak, protein = self.get_protein_for_expression(), **keywords)
        
        return species

    def update_reactions(self, **keywords):
        mech_tx = self.get_mechanism("transcription")
        
        reactions = [] #a list of reactions must be returned

        reactions += mech_tx.update_reactions(dna = self.dna_to_bind, transcript = self.transcript, regulator = self.repressor, 
                                             component = self, part_id = self.name+"_"+self.repressor.name, leak = self.leak, protein = self.get_protein_for_expression(), **keywords)
        return reactions


class CombinatorialPromoter(Promoter):
    def __init__(self, name, regulators, leak = False, assembly = None,
                 transcript = None, length = 0, mechanisms = None,
                 parameters = None,protein=None,tx_capable_list = None,cooperativity = None, **keywords):
        """
        A combinatorial promoter is something where binding multiple regulators result in
        qualitatively different transcription behaviour. For example, maybe it's an AND
        gate promoter where it only transcribes if two regulators are bound, but not if 
        either one is bound.
        
        =============
        inputs
        =============
        name: the name of the promoter
        regulators: a list of strings or species indicating all the possible regualtors that can bind
        
        leak: if true, then a promoter with nothing bound will transcribe
        
        assembly: a DNA_assembly object that contains this promoter
        
        transcript: the transcript that this promoter makes
        
        length: the length in nt? I don't think this is used for anything at the moment
        
        mechanisms: additional mechanisms. formatted with {"mechanism_type":mechanismObject(),...}
        
        parameters: promoter-specific parameters. Formatted as {("identifier1","identifier2"):value,...}
        
        tx_capable_list: list of which combination of regulators bound will lead to transcription. 
                        formatted as [["regulator1","regulator2"],["regulator1"],...] regulators
                        can be strings or Species
        cooperativity: a dictionary of cooperativity values. For example, {"regulator":2,"regulator2":1,....}
        """

        Promoter.__init__(self, name = name, assembly = assembly,
                          transcript = transcript, length = length,
                          mechanisms = mechanisms, parameters = parameters, protein=protein,
                          **keywords)

        if not isinstance(regulators, list):
            #you could give one string as a regulator
            regulators = [regulators]
        self.cooperativity = cooperativity
        self.regulators = []
        for regulator in regulators:
            self.regulators += [self.set_species(regulator, material_type = "protein")]
            
        #after we've sanitized the inputs, then sort
        self.regulators = sorted(self.regulators)
        #now let's work out the tx_capable_list
        if tx_capable_list is None:
            #if nothing is passed, that means everything transcribes
            allcomb = []
            for r in range(1,len(self.regulators)+1):
                #make all combinations of regulators
                allcomb += [set(a) for a in it.combinations([a.name for a in self.regulators],r)]
            self.tx_capable_list = allcomb
        elif isinstance(tx_capable_list, list):
            #if the user passed a list then the user knows what they want
            newlist = []
            #this part converts any species in the tx_capable_list into a string
            for element in tx_capable_list:
                sublist = []
                for specie in element:
                    if(isinstance(specie,Species)):
                        sublist += [specie.name]
                    else:
                        sublist += [specie]
                newlist+=[sublist]
            self.tx_capable_list = [set(a) for a in newlist]

        self.leak = leak
        self.complex_combinations = {}
        self.tx_capable_complexes = []
        self.leak_complexes = []
        self.add_mechanism(Combinatorial_Cooperative_Binding(), "binding", overwrite = True)
        
        

    def update_species(self):

        mech_tx = self.get_mechanism("transcription")
        mech_b = self.get_mechanism('binding')
        #set the tx_capable_complexes to nothing because we havent updated species yet!
        self.tx_capable_complexes = []
        self.leak_complexes = []
        species = [self.dna_to_bind]+self.regulators
        self.complexes = []
        bound_species = mech_b.update_species(self.regulators,self.dna_to_bind,\
                        component = self,part_id = self.name,cooperativity=self.cooperativity,protein=self.protein)
        #above is all the species with DNA bound to regulators. Now, we need to extract only the ones which
        #are transcribable

        if self.leak is not False:
            #this part takes care of the promoter not bound to anything
            species += mech_tx.update_species(dna = self.dna_to_bind, transcript = self.transcript, protein = self.get_protein_for_expression(), component = self, part_id = self.name+"_leak")
            #self.leak_complexes += []

        for bound_complex in bound_species: 
            species_inside = []
            for regulator in self.regulators:
                if(bound_complex.contains_species_monomer(regulator) and bound_complex.contains_species_monomer(self.dna_to_bind)):
                    species_inside += [regulator.name]
            if(set(species_inside) in [set(a) for a in self.tx_capable_list]):
                #only the transcribable complexes in tx_capable_list get transcription reactions
                tx_capable_species = mech_tx.update_species(dna = bound_complex, transcript = self.transcript, \
                                                protein = self.get_protein_for_expression(), component = self, part_id = self.name)
                species +=tx_capable_species
                self.tx_capable_complexes +=[bound_complex]
            else:
                #in this case there's a combination of regulators which does not feature in tx_capable_list
                #this means: 
                # 1) this complex does nothing so don't add it
                # 2) we said we wanted leak, so then you should add this, but with the "_leak" parameters
                #                                                         (that happens in update_reactions)
                if(self.leak is not False):
                    leak_species = mech_tx.update_species(dna = bound_complex, transcript = self.transcript, \
                                        protein = self.get_protein_for_expression(), component = self, part_id = self.name+"_leak")
                    species += leak_species
                    self.leak_complexes += [bound_complex]
        species+=bound_species  
        return species

    def update_reactions(self):
        
        reactions = []
        mech_tx = self.get_mechanism("transcription")
        mech_b = self.get_mechanism('binding')

        if self.leak is not False:
            #once again the DNA not bound to anything gets special treatment
            reactions += mech_tx.update_reactions(dna = self.dna_to_bind, component = self, part_id = self.name+"_leak", \
                                                            transcript = self.transcript, protein = self.protein)
        #the binding reactions happen no matter what
        reactions += mech_b.update_reactions(self.regulators,self.dna_to_bind,component = self,\
                                                        part_id = self.name,cooperativity=self.cooperativity, protein = self.protein)
        if((self.tx_capable_complexes is None) or self.tx_capable_complexes == []):
            species = self.update_species()
            #this could mean we haven't run update_species() yet
            
            if(self.tx_capable_complexes == []):
                if(self.leak_complexes is None or self.leak_complexes == []):
                    #if it's still zero after running update_species then we could be in trouble
                    warn("nothing can transcribe from combinatorial promoter {}".format(self.name))
            
        if(len(self.tx_capable_complexes)>0):
            for specie in self.tx_capable_complexes:
                tx_partid = self.name
                for part in specie.species_set:
                    #construct the name of the promoter with regulators bound
                    if part.material_type == "dna":
                        #the DNA doesn't matter
                        pass
                    else:
                        #put in the regulators!
                        tx_partid += "_"+part.name
                if(tx_partid[0]=="_"):
                    #this will only happen if the name of the dna is ""
                    tx_partid = tx_partid[1:]
                #if it's bound to RNAP then it transcribes, right?
                tx_partid = tx_partid+"_RNAP"
                reactions += mech_tx.update_reactions(dna = specie, component = self, part_id = tx_partid, \
                                            transcript = self.transcript, protein = self.get_protein_for_expression())
        if(len(self.leak_complexes)>0):
            for specie in self.leak_complexes:
                #in this case every reaction uses the "promoter_leak" partid
                leak_partid = self.name+"_leak"
                reactions += mech_tx.update_reactions(dna = specie, component = self, part_id = leak_partid, \
                                            transcript = self.transcript, protein = self.get_protein_for_expression())

        return reactions



# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.


from .mechanism import Mechanism
from .reaction import Reaction
from .species import Complex
from .component import Component
from .propensities import ProportionalHillPositive



class BasicIntegration(Mechanism):
    """Mechanism for the schema DNA1 + DNA2 --> DNA3 + DNA4."""
    def __init__(self, name: str="basic_integration", mechanism_type: str="integration",**keywords):
        """Initializes a BasicIntegration instance.

        :param name: name of the Mechanism, default: basic_integration
        :param mechanism_type: type of the Mechanism, default: integration
        :param keywords:
        """
        Mechanism.__init__(self, name, mechanism_type)

    def update_species(self, DNA_inputs, DNA_outputs = None, **keywords):
        #this doesn't make any species because I use a Binding mechanism for that
        #maybe if we do the tetramerization mechanism then this would do something
        return []

    def update_reactions(self, DNA_inputs, DNA_outputs, component = None, part_id = None, kint = None, **keywords):
        if part_id is None and component is not None:
            part_id = component.name

        if kint is None and component is None:
            raise ValueError("Must pass in either a component or kint.")
        elif kint is None:
            kint = component.get_parameter("kint", part_id = part_id, mechanism = self)

        return [Reaction.from_massaction(inputs=DNA_inputs, outputs=DNA_outputs, k_forward=kint)]

class EnzymeIntegration(Mechanism):
    """Mechanism for the schema integrase+DNA1 + DNA2 --> integrase+DNA3 + DNA4."""
    def __init__(self, name: str="enzyme_integration", mechanism_type: str="integration",integrase="Int1",**keywords):
        """Initializes a BasicIntegration instance.

        :param name: name of the Mechanism, default: basic_integration
        :param mechanism_type: type of the Mechanism, default: integration
        :param keywords:
        """
        # TODO ZAT: remove unused keywords argument
        self.integrase = Component.set_species(integrase,material_type="protein")
        Mechanism.__init__(self, name, mechanism_type)

    def update_species(self, DNA_inputs, DNA_outputs = None, **keywords):
        #this doesn't make any species because I use a Binding mechanism for that
        #maybe if we do the tetramerization mechanism then this would do something
        return []

    def update_reactions(self, DNA_inputs, DNA_outputs, component = None, part_id = None, kint = None, **keywords):
        
        
        if part_id is None and component is not None:
            part_id = component.name

        if kint is None and component is None:
            raise ValueError("Must pass in either a component or kint.")
        
        elif kint is None:
            kint = component.get_parameter("kint", part_id = part_id, mechanism = self)

        return [Reaction.from_massaction(inputs=[self.integrase]*4+DNA_inputs,\
                                        outputs=[self.integrase]*4+DNA_outputs, k_forward=kint)]



#  Copyright (c) 2019, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.

import copy
from numbers import Real
from typing import List, Union
from warnings import warn

from .global_mechanism import GlobalMechanism
from .mechanism import Mechanism
from .parameter import Parameter, ParameterDatabase, ParameterKey
from .species import Species


class Component(object):
    """Component class for core components.

    These subclasses of Component represent different kinds of biomolecules.

    This class must be Subclassed to provide functionality with the functions get_species and get_reactions overwritten.
    """
    def __init__(self, name: Union[str, Species],
                 mechanisms=None,  # custom mechanisms
                 parameters=None,  # parameter configuration
                 parameter_file=None,  # custom parameter file
                 mixture=None,
                 attributes=None,
                 initial_concentration = None, #This is added as a parameter ("initial concentration", None, self.name):initial_concentration
                 initial_condition_dictionary=None,
                 **keywords  # parameter keywords
                 ):
        """Initializes a Component object.

        :param name:
        :param mechanisms:
        :param parameters:
        :param parameter_file:
        :param mixture:
        :param attributes:
        :param initial_concentration:
        :param initial_condition_dictionary:
        :param keywords:
        """
        if mechanisms is None:
            self.mechanisms = {}
        else:
            self.mechanisms = mechanisms
        if isinstance(name, Species):
            self.name = name.name
        elif isinstance(name, str):
            self.name = name
        else:
            raise ValueError("name must be a Species or string")

        # Attributes can be used to store key words like protein deg-tags for
        # components that mimic CRN species.
        self.attributes = []
        self.set_attributes(attributes)

        if mixture is not None:
            self.set_mixture(mixture)
        else:
            self.set_mixture(None)

        self.parameter_database = ParameterDatabase(parameter_file=parameter_file, parameter_dictionary=parameters)
        self.initial_concentration = initial_concentration

        # Components can also store initial conditions, just like Mixtures
        if initial_condition_dictionary is None:
            self.initial_condition_dictionary = {}
        else:
            self.initial_condition_dictionary = dict(initial_condition_dictionary)
    @property
    def initial_concentration(self):
        return self._initial_concentration

    @initial_concentration.setter
    def initial_concentration(self, initial_concentration):
        if initial_concentration is not None:
            if initial_concentration < 0:
                raise ValueError(f'Initial concentration must be non-negative, this was given: {initial_concentration}')
            param_key = ParameterKey(mechanism = "initial concentration", part_id = None, name = self.name)
            self.parameter_database.add_parameter(parameter_name = "initial concentration", parameter_value = initial_concentration, parameter_key = param_key, overwrite_parameters = True)
    
        self._initial_concentration = initial_concentration

    def set_mixture(self, mixture) -> None:
        """Set the mixture the Component is in.

        :param mixture:
        :return: None
        """
        self.mixture = mixture

    # TODO implement as an abstractmethod
    def get_species(self) -> None:
        """The subclasses should implement this method!

        :return: None
        """
        return None
    @classmethod
    def set_species(self, species: Union[Species, str], material_type=None, attributes=None) -> Species:
        """Helper function that allows species to be set from strings, species, or Components

        :param species: Species, str, Component
        :param material_type:
        :param attributes:
        :return: Species
        """
        if isinstance(species, Species):
            return species
        elif isinstance(species, str):
            return Species(name=species, material_type=material_type, attributes=attributes)
        elif isinstance(species, Component) and species.get_species() is not None:
            return species.get_species()
        elif isinstance(species, list):
            return [self.set_species(s, material_type = material_type, attributes = attributes) for s in species]
        else:
            raise ValueError(f"Invalid Species: string, chemical_reaction_network.Species or Component with implemented .get_species() required as input. Recieved {species}")

    def __hash__(self):
        return str.__hash__(repr(self.get_species()))

    def set_attributes(self, attributes: List[str]):
        if attributes is not None:
            for attribute in attributes:
                self.add_attribute(attribute)

    def add_attribute(self, attribute: str):
        assert isinstance(attribute, str) and attribute is not None, "Attribute: %s must be a str" % attribute

        self.attributes.append(attribute)
        if hasattr(self, 'species') and self.species is not None:
            self.species.add_attribute(attribute)
        else:
            raise Warning(f"Component {self.name} has no internal species and therefore no attributes")

    def update_parameters(self, parameter_file = None, parameters = None, parameter_database = None, overwrite_parameters = True):
        """Updates the ParameterDatabase inside a Component

        Possible inputs:
            parameter_file (string)
            parameters (dict)
            parameter_database (ParameterDatabase)        
        """

        if parameter_file is not None:
            self.parameter_database.load_parameters_from_file(parameter_file, overwrite_parameters = overwrite_parameters)
            
        if parameters is not None:
            self.parameter_database.load_parameters_from_dictionary(parameters, overwrite_parameters = overwrite_parameters)

        if parameter_database is not None:
            self.parameter_database.load_parameters_from_database(parameter_database, overwrite_parameters = overwrite_parameters)

    def get_mechanism(self, mechanism_type, optional_mechanism=False):
        """Searches the Component for a Mechanism of the correct type.

        If the Component does not have the mechanism, searches the Components' Mixture for the Mechanism.

        :param mechanism_type:
        :param optional_mechanism: toggles whether an error is thrown if no mechanism is found
        :return:
        """
        if not isinstance(mechanism_type, str):
            raise TypeError(f"mechanism_type must be a string. Received {mechanism_type}.")

        mech = None
        if mechanism_type in self.mechanisms:
            return self.mechanisms[mechanism_type]
        elif self.mixture is not None:
            mech = self.mixture.get_mechanism(mechanism_type)
        if mech is None and not optional_mechanism:
            raise KeyError(f"Unable to find mechanism of type {mechanism_type} in Component {self} or Mixture {self.mixture}.")
        else:
            return mech

    @property
    def mechanisms(self):
        return self._mechanisms

    @mechanisms.setter
    def mechanisms(self, mechanisms):
        self._mechanisms = {}
        if isinstance(mechanisms, dict):
            for mech_type in mechanisms:
                self.add_mechanism(mechanisms[mech_type], mech_type, overwrite = True)
        elif isinstance(mechanisms, list):
            for mech in mechanisms:
                self.add_mechanism(mech, overwrite = True)

    def add_mechanism(self, mechanism: Mechanism, mech_type=None, overwrite=False, optional_mechanism=False):
        """adds a mechanism of type mech_type to the Component Mechanism dictionary.

        :param mechanism:
        :param mech_type:
        :param overwrite: toggles whether the mechanism is added overwriting any mechanism with the same key.
        :param optional_mechanism: toggles whether an error is thrown if a Mechanism is added that conflicts with an exising Mechanism
        :return:
        """
        if not hasattr(self, "_mechanisms"):
            self._mechanisms = {}
            
        if not isinstance(mechanism, Mechanism):
            raise TypeError(f"mechanism must be a Mechanism. Received {mechanism}.")

        if mech_type is None:
            mech_type = mechanism.mechanism_type
        if not isinstance(mech_type, str):
            raise TypeError(f"mechanism keys must be strings. Received {mech_type}")
            
        if mech_type not in self._mechanisms or overwrite:
            self._mechanisms[mech_type] = copy.deepcopy(mechanism)
        elif not optional_mechanism:
            raise ValueError(f"mech_type {mech_type} already in component {self}. To overwrite, use keyword overwrite = True.")

    def add_mechanisms(self, mechanisms: Union[Mechanism, GlobalMechanism], overwrite=False, optional_mechanism=False):
        """This function adds a list or dictionary of mechanisms to the mixture.

        :param mechanisms: Can take both GlobalMechanisms and Mechanisms
        :param overwrite: toggles whether the mechanism is added overwriting any mechanism with the same key.
        :param optional_mechanism: toggles whether an error is thrown if a Mechanism is added that conflicts with an exising Mechanism
        :return:
        """

        if isinstance(mechanisms, Mechanism):
            self.add_mechanism(mechanisms, overwrite=overwrite, optional_mechanism=optional_mechanism)
        elif isinstance(mechanisms, dict):
            for mech_type in mechanisms:
                self.add_mechanism(mechanisms[mech_type], mech_type, overwrite=overwrite, optional_mechanism=optional_mechanism)
        elif isinstance(mechanisms, list):
            for mech in mechanisms:
                self.add_mechanism(mech, overwrite=overwrite, optional_mechanism=optional_mechanism)
        else:
            raise ValueError(f"add_mechanisms expected a list of Mechanisms. Received {mechanisms}")

    def get_parameter(self, param_name: str, part_id=None, mechanism=None, return_numerical=False, return_none = False, check_mixture = True) -> Union[Parameter, Real]:
        """Get a parameter from different objects that hold parameters.

        Hierarchy:
         1. tries to find the Parameter in Component.parameter_database
         2. tries to find the parameter in Component.mixture.parameter_database

        :param param_name:
        :param part_id:
        :param mechanism:
        :param return_numerical: numerical value or the parameter object is returned
        :param return_none: returns None instead of throwing an error if a parameter isn't found
        :param check_mixture: toggle whether or not to check the Component's Mixture as well
        :return: Parameter object or a Real number
        """
        # Try the Component ParameterDatabase
        param = self.parameter_database.find_parameter(mechanism, part_id, param_name)

        # Next try the Mixture ParameterDatabase
        if param is None and self.mixture is not None and check_mixture:
            param = self.mixture.get_parameter(mechanism, part_id, param_name)

        if param is None and not return_none:
            raise ValueError("No parameters can be found that match the "
                             "(mechanism, part_id, "
                             f"param_name)=({repr(mechanism)}, {part_id}, "
                             f"{param_name})")
        elif return_numerical and isinstance(param, Parameter):
            return param.value
        else:
            return param

    # TODO implement abstractmethod
    def update_species(self) -> List[Species]:
        """The subclasses should implement this method!

        :return: empty list
        """
        species = []
        warn("Unsubclassed update_species called for " + repr(self))
        return species

    # TODO implement abstractmethod
    def update_reactions(self) -> List[Species]:
        """The subclasses should implement this method!

        :return: empty list
        """
        reactions = []
        warn("Unsubclassed update_reactions called for " + repr(self))
        return reactions
        
    def enumerate_components(self,**keywords) -> List:
        """this is for component enumeration. Usually you will return a list of components that are
        copies of existing ones (first list) and new components (second list). For example,
        A DNA_construct makes a list of copies of its parts as the first output, and a list of RNA_constructs
        as the second output.
        An RNA_construct will make a list of copies of its parts as the first output, and a list of Protein
        components as its second output (if it makes any proteins)"""
        return []


    def __repr__(self):
        return type(self).__name__ + ": " + self.name



from setuptools import setup

# Get the long description from the README file
with open('README.md') as fp:
    long_description = fp.read()

setup(
    name='biocrnpyler',
    version='1.1.2',
    author='Build-A-Cell',
    url='https://github.com/BuildACell/biocrnpyler/',
    description='A chemical reaction network compiler for generating large biological circuit models',
    long_description=long_description,
    packages=['biocrnpyler'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Operating System :: OS Independent',
    ],
    install_requires=[
          "python-libsbml",
          ],
    extras_require = { 
        "all": [
            "numpy<2.0",
            "matplotlib",
            "networkx",
            "bokeh>=1.4.0",
            "fa2_modified",
            "jupyter",
            "pytest",
            "pytest-cov",
            "nbconvert"
            ],
        },
    tests_require=["pytest", "pytest-cov"],
    setup_requires=["pytest-runner"],
    python_requires='>=3.6',
    keywords="SBML synthetic systems biology modeling Chemical Reaction Network CRN model",
    project_urls={
    'Documentation': 'https://readthedocs.org/projects/biocrnpyler/',
    'Funding': 'http://www.cds.caltech.edu/~murray/wiki/index.php?title=Developing_Standardized_Cell-Free_Platforms_for_Rapid_Prototyping_of_Synthetic_Biology_Circuits_and_Pathways',
    'Source': 'https://github.com/BuildACell/biocrnpyler',
    'Tracker': 'https://github.com/BuildACell/BioCRNPyler/issues',
    }, # give an installation message about bioscrape / roadrunner.
)



#  Copyright (c) 2019, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.

from typing import List, Union

from .component import Component
from .components_basic import DNA, RNA, Protein
from .dna_part_promoter import Promoter
from .dna_part_rbs import RBS
from .mechanism import Mechanism
from .mixture import Mixture
from .parameter import ParameterDatabase
from .reaction import Reaction
from .species import Species


class DNAassembly(DNA):
    """A class that contains a Promoter, RBS, transcript, and protein."""

    def __init__(self, name: str, dna=None, promoter=None, transcript=None,
                 rbs=None, protein=None, length=None,
                 attributes=None, mechanisms=None, parameters=None, initial_concentration=None, **keywords):
        """
            Note:
        If transcript is None and protein is not None,
        the DNAassembly will use its transcription mechanisms to produce the protein.
        This is used by Expression Mixtures.
        :param name: name of the DNA assembly
        :param dna:
        :param promoter:
        :param transcript:
        :param rbs:
        :param protein:
        :param length:
        :param attributes:
        :param mechanisms:
        :param parameters:
        :param initial_concentration:
        :param keywords: passed into the parent object (DNA)
        """
        self.promoter = None
        self.rbs = None
        self.transcript = None
        
        # This has to be called at the end so mechanisms are set for the promoter, RBS, etc.
        DNA.__init__(self, name, length=length, mechanisms=mechanisms,
                     parameters=parameters, initial_concentration=initial_concentration,
                     attributes=attributes, **keywords)

        self.update_dna(dna, attributes=attributes)
        self.update_transcript(transcript)
        self.update_protein(protein)
        self.update_promoter(promoter, transcript=self.transcript, protein=self.protein)
        self.update_rbs(rbs, transcript=self.transcript, protein=self.protein)

    def get_species(self):
        return self.dna
        
    def set_mixture(self, mixture: Mixture) -> None:
        """Set the mixture the Component is in.

        :param mixture: reference to a Mixture instance
        :return: None
        """
        self.mixture = mixture
        if self.promoter is not None:
            self.promoter.set_mixture(mixture)
        if self.rbs is not None:
            self.rbs.set_mixture(mixture)

    def update_dna(self, dna: Union[None, DNA, str], attributes=None) -> None:
        """sets up the dna attribute with a valid DNA instance.

        :param dna: name of a dna sequence or a DNA instance
        :param attributes: Species attribute
        :return: None
        """
        if dna is None:
            self.dna = self.set_species(self.name, material_type="dna", attributes=attributes)
        else:
            self.dna = self.set_species(dna, material_type="dna", attributes=attributes)

        if self.promoter is not None:
            self.promoter.dna = self.dna
        if self.rbs is not None:
            self.rbs.dna = self.dna
        
    def update_transcript(self, transcript: Union[None, RNA, str, bool], attributes=None) -> None:
        """sets up the transcript attribute with a valid RNA instance.

        :param transcript: name of a RNA transcript or RNA instance
        :param attributes: Species attribute
        :return: None
        """
        if transcript is None:
            self.transcript = self.set_species(self.name, material_type="rna", attributes=attributes)

        # this is used for expression mixtures where there is no transcript!
        elif transcript is False:
            self.transcript = None
        else:
            self.transcript = self.set_species(transcript, material_type="rna", attributes=attributes)

        if self.promoter is not None:
            self.promoter.transcript = self.transcript
        if self.rbs is not None:
            self.rbs.transcript = self.transcript

    def update_protein(self, protein: Union[None, Protein, str], attributes=None) -> None:
        """sets up the protein attribute with a valid Protein instance.

        :param protein: name of a protein or Protein instance
        :param attributes: Species attribute
        :return: None
        """
        if protein is None:
            self.protein = self.set_species(self.name, material_type="protein", attributes=attributes)
        else:
            self.protein = self.set_species(protein, material_type="protein", attributes=attributes)

        if self.rbs is not None:
            self.rbs.protein = self.protein
        if self.promoter is not None:
            self.promoter.protein = self.protein

    def update_promoter(self, promoter: Union[Protein, str], transcript: RNA=None, protein: Protein=None) -> None:
        """sets up the promoter attribute with a valid Promoter instance.

        :param promoter: name of a promoter or Promoter instance
        :param transcript: reference to the RNA transcript
        :param protein:
        :return: None
        """
        if transcript is not None:
            self.update_transcript(transcript)

        if promoter is not None:
            self.promoter = Promoter.from_promoter(name=promoter, assembly=self,
                                                   transcript=self.transcript,
                                                   protein=protein)
        else:
            self.promoter = None

        if self.promoter is not None:
            self.promoter.update_parameters(parameter_database=self.parameter_database, overwrite_parameters=False)
            self.promoter.set_mixture(self.mixture)
            self.promoter.add_mechanisms(self.mechanisms, optional_mechanism=True)

    def update_rbs(self, rbs: Union[RBS, str], transcript: RNA=None, protein: Protein=None) -> None:
        """sets up the rbs attribute with a valid RBS instance.

        :param rbs: name of the ribosome binding site or RBS instance
        :param transcript: RNA that contains the ribosome binding site
        :param protein: protein that RNA contains
        :return: None
        """
        if protein is not None:
            self.update_protein(protein)

        if transcript is not None:
            self.update_transcript(transcript)

        if rbs is not None:
            self.rbs = RBS.from_rbs(name=rbs, assembly=self, protein=self.protein,
                                    transcript=self.transcript)
        else:
            self.rbs = None

        if self.rbs is not None:
            self.rbs.update_parameters(parameter_database=self.parameter_database, overwrite_parameters=False)
            self.rbs.set_mixture(self.mixture)
            self.rbs.add_mechanisms(self.mechanisms, optional_mechanism=True)

    def update_species(self) -> List[Species]:
        """collects the list of Species that a DNAassemlby instance holds.

        :return: list of Species that a DNAassemlby instance holds
        """
        species = []
        species.append(self.dna)
        if self.promoter is not None:
            species += self.promoter.update_species()

        if self.rbs is not None:
            species += self.rbs.update_species()

        #deg_mech = self.get_mechanism("rna_degredation", optional_mechanism=True)
        #if deg_mech is not None and self.promoter is not None and self.transcript is not None:
        #    species += deg_mech.update_species(rna=self.transcript, component=self.promoter, part_id=self.transcript.name)

        return species

    def update_reactions(self) -> List[Reaction]:
        """collects the list of Reactions that a DNAassemlby instance holds.

        :return: list of Reactions that a DNAassemlby instance holds.
        """
        reactions = []
        if self.promoter is not None:
            reactions += self.promoter.update_reactions()

        if self.rbs is not None:
            reactions += self.rbs.update_reactions()

        #deg_mech = self.get_mechanism("rna_degredation", optional_mechanism=True)
        #if deg_mech is not None and self.promoter is not None and self.transcript is not None:
        #    reactions += deg_mech.update_reactions(rna=self.transcript, component=self.promoter, part_id=self.transcript.name)

        return reactions

    def update_parameters(self, parameter_file: str=None, parameters:ParameterDatabase=None, overwrite_parameters: bool=True) -> None:
        """updates the parameters stored in dna, promoter and rbs.

        :param parameter_file: valid parameter file
        :param parameters: a parameter database instance
        :param overwrite_parameters: whether to overwrite existing parameters
        :return: None
        """

        DNA.update_parameters(self=self, parameter_file=parameter_file, parameters=parameters, overwrite_parameters=overwrite_parameters)

        if self.promoter is not None:
            self.promoter.update_parameters(parameter_file=parameter_file, parameters=parameters, overwrite_parameters=overwrite_parameters)

        if self.rbs is not None:
            self.rbs.update_parameters(parameter_file=parameter_file, parameters=parameters, overwrite_parameters=overwrite_parameters)

    def add_mechanism(self, mechanism: Mechanism, mech_type: str=None, overwrite: bool=False, optional_mechanism: bool=False) -> None:
        """adds a mechanism of type mech_type to the Component Mechanism dictionary.

        DNA_assembly also adds the mechanisms to its promoter and rbs (but never overwrites them!)

        :param mechanism: reference to a Mechanism instance
        :param mech_type: type of mechanism
        :param overwrite: whether to overwrite the mechanism in Component
        :param optional_mechanism:
        :return: None
        """
        Component.add_mechanism(self, mechanism, mech_type=mech_type, overwrite=overwrite, optional_mechanism=optional_mechanism)

        if self.promoter is not None:
            self.promoter.add_mechanism(mechanism, mech_type=mech_type, optional_mechanism=True)

        if self.rbs is not None:
            self.rbs.add_mechanism(mechanism, mech_type=mech_type, optional_mechanism=True)

    def __str__(self):
        return type(self).__name__ + ": " + self.name

    def __repr__(self):
        txt = str(self)
        if self.promoter is not None:
            txt += "\n\t" + repr(self.promoter)
            txt += "\n\ttranscript = " + repr(self.transcript)
        if self.rbs is not None:
            txt += "\n\t" + repr(self.rbs)
            txt += "\n\tprotein = " + repr(self.protein)
        return txt




# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

from .chemical_reaction_network import ChemicalReactionNetwork
from .components_basic import Protein, Metabolite
from .dna_assembly import DNAassembly
from .global_mechanism import Degredation_mRNA_MM, Dilution
from .mechanism import EmptyMechanism
from .mechanisms_binding import One_Step_Binding
from .mechanisms_enzyme import BasicCatalysis, MichaelisMenten
from .mechanisms_txtl import (OneStepGeneExpression, SimpleTranscription,
                              SimpleTranslation, Transcription_MM, Translation_MM,
                              Energy_Transcription_MM, Energy_Translation_MM)
from .mechanisms_metabolite import OneStepPathway
from .mixture import Mixture


class ExpressionExtract(Mixture):
    """A Model for Gene Expression without any Machinery (eg Ribosomes, Polymerases, etc.).

    Here transcription and Translation are lumped into one reaction: expression.
    """
    def __init__(self, name="", **kwargs):
        """Initializes an ExpressionExtract instance.

        :param name: name of the mixture
        :param kwargs: keywords passed into the parent Class (Mixture)
        """
        # always call the superlcass Mixture.__init__(...)
        Mixture.__init__(self, name=name, **kwargs)

        # Create default Expression Mechanisms
        dummy_translation = EmptyMechanism(name="dummy_translation", mechanism_type="translation")
        mech_expression = OneStepGeneExpression()
        mech_cat = BasicCatalysis()
        mech_bind = One_Step_Binding()

        default_mechanisms = {
            mech_expression.mechanism_type: mech_expression,
            dummy_translation.mechanism_type: dummy_translation,
            mech_cat.mechanism_type: mech_cat,
            mech_bind.mechanism_type: mech_bind
        }

        self.add_mechanisms(default_mechanisms)

    def compile_crn(self, **keywords) -> ChemicalReactionNetwork:
        """Overwriting compile_crn to turn off transcription in all DNAassemblies

        :return: compiled CRN instance
        """
        for component in self.components:
            if isinstance(component, DNAassembly):
                # Only turn off transcription for an Assembly that makes a Protein.
                # Some assemblies might only make RNA!
                if component.protein is not None:
                    # This will turn off transcription and set Promoter.transcript = False
                    # Mechanisms that recieve no transcript but a protein will use the protein instead.
                    component.update_transcript(False)

        # Call the superclass function
        return Mixture.compile_crn(self, **keywords)


class SimpleTxTlExtract(Mixture):
    """
    A Model for Transcription and Translation in an extract any Machinery (eg Ribosomes, Polymerases, etc.)
    RNA is degraded via a global mechanism.
    """

    def __init__(self, name="", **kwargs):
        """Initializes a SimpleTxTlExtract instance.

        :param name: name of the mixture
        :param kwargs: keywords passed into the parent Class (Mixture)
        """
        # Always call the superlcass Mixture.__init__(...)
        Mixture.__init__(self, name=name, **kwargs)

        # TxTl Mechanisms
        mech_tx = SimpleTranscription()
        mech_tl = SimpleTranslation()
        mech_cat = BasicCatalysis()
        mech_bind = One_Step_Binding()

        default_mechanisms = {
            mech_tx.mechanism_type: mech_tx,
            mech_tl.mechanism_type: mech_tl,
            mech_cat.mechanism_type: mech_cat,
            mech_bind.mechanism_type: mech_bind
        }
        self.add_mechanisms(default_mechanisms)

        # global mechanisms for dilution and rna degredation
        mech_rna_deg_global = Dilution(name="rna_degredation", filter_dict={"rna": True}, default_on=False)
        global_mechanisms = {"rna_degredation": mech_rna_deg_global}
        self.add_mechanisms(global_mechanisms)        


class TxTlExtract(Mixture):
    """A Model for Transcription and Translation in Cell Extract with Ribosomes, Polymerases, and Endonucleases.

    This model does not include any energy
    """
    def __init__(self, name="", rnap="RNAP", ribosome="Ribo", rnaase="RNAase", **kwargs):
        """Initializes a TxTlExtract instance.

        :param name: name of the mixture
        :param rnap: name of the RNA polymerase, default: RNAP
        :param ribosome: name of the ribosome, default: Ribo
        :param rnaase: name of the Ribonuclease, default: RNAase
        :param kwargs: keywords passed into the parent Class (Mixture)
        """
        # Always call the superlcass Mixture.__init__(...)
        Mixture.__init__(self, name=name, **kwargs)
        
        # create default Components to represent cellular machinery
        self.rnap = Protein(rnap)
        self.ribosome = Protein(ribosome)
        self.rnaase = Protein(rnaase)

        default_components = [self.rnap, self.ribosome, self.rnaase]
        self.add_components(default_components)

        # Create default TxTl Mechanisms
        mech_tx = Transcription_MM(rnap=self.rnap.get_species())
        mech_tl = Translation_MM(ribosome=self.ribosome.get_species())
        mech_rna_deg = Degredation_mRNA_MM(nuclease=self.rnaase.get_species())
        mech_cat = MichaelisMenten()
        mech_bind = One_Step_Binding()

        default_mechanisms = {
            mech_tx.mechanism_type: mech_tx,
            mech_tl.mechanism_type: mech_tl,
            mech_rna_deg.mechanism_type: mech_rna_deg,
            mech_cat.mechanism_type: mech_cat,
            mech_bind.mechanism_type: mech_bind
        }
        self.add_mechanisms(default_mechanisms)



class EnergyTxTlExtract(Mixture):
    """A Model for Transcription and Translation in Cell Extract with Ribosomes, Polymerases, and Endonucleases.

    This model include energy carrier molcules in the form of NTPs, Amino Acids, and a Fuel Species (such as 3PGA) used for NTP
    regeneration. This model is equivalent to TxTl extract, but with limited fuel. Note that different amino acids and nucleotides
    are lumped together.

    Energy usage for transcription and translation is length dependent."""

    def __init__(self, name="", rnap="RNAP", ribosome="Ribo", rnaase="RNAase", 
        ntps = "NTPs", ndps = "NDPs", amino_acids = "amino_acids", fuel = "Fuel_3PGA", **kwargs):
        """
        :param name: name of the mixture
        :param rnap: name of the RNA polymerase, default: RNAP
        :param ribosome: name of the ribosome, default: Ribo
        :param rnaase: name of the Ribonuclease, default: RNAase
        :param ntps: name of the nucleotide fuel source (eg ATP + GTP etc), default: NTP
        :param amino_acids: name of the amino acids species, default: amino_acids
        :param fuel: name of the fuel species that regenerates ATP
        :param kwargs: keywords passed into the parent Class (Mixture)
        """
        Mixture.__init__(self, name=name, **kwargs)
        
        # create default Components to represent cellular machinery
        self.rnap = Protein(rnap)
        self.ribosome = Protein(ribosome)
        self.rnaase = Protein(rnaase)
        self.amino_acids = Metabolite(amino_acids)
        self.fuel = Metabolite(fuel) #fuel is degraded into things other than ATP as well
        self.ndps = Metabolite(ndps) #NDPs
        self.ntps = Metabolite(ntps, precursors = [self.fuel, self.ndps], products = [self.ndps]) #fuel becomes ATP, and ATP is degraded
        

        #These mechanisms are Component specific and only added to the NTPs metabolite
        mech_pathway = OneStepPathway()
        self.ntps.add_mechanisms(mech_pathway)
        self.fuel.add_mechanisms(mech_pathway)


        default_components = [self.rnap, self.ribosome, self.rnaase, self.amino_acids, self.ntps, self.fuel]
        self.add_components(default_components)

        # Create default TxTl Mechanisms
        mech_tx = Energy_Transcription_MM(rnap=self.rnap.get_species(), fuels = [self.ntps.get_species()], wastes = [])
        mech_tl = Energy_Translation_MM(ribosome=self.ribosome.get_species(), fuels = 4*[self.ntps.get_species()] +[self.amino_acids.get_species()], wastes = 4*[self.ndps.get_species()])
        mech_rna_deg = Degredation_mRNA_MM(nuclease=self.rnaase.get_species())
        mech_cat = MichaelisMenten()
        mech_bind = One_Step_Binding()

        default_mechanisms = {
            mech_tx.mechanism_type: mech_tx,
            mech_tl.mechanism_type: mech_tl,
            mech_rna_deg.mechanism_type: mech_rna_deg,
            mech_cat.mechanism_type: mech_cat,
            mech_bind.mechanism_type: mech_bind
        }
        self.add_mechanisms(default_mechanisms)



